import numpy as np
from scipy import sparse as sps
from tqdm import tqdm
from ._lda import LabelledLDATrainer
from .lda import (
    Predictor,
    RealType, IntegerType, IndexType,
    number_to_array, check_array,
    LDAPredictorMixin
)


class LabelledLDA(LDAPredictorMixin):
    def __init__(self,
                 alpha=1e-2, epsilon=1e-30, topic_word_prior=None, add_dummy_topic=False,
                 n_iter=1000,
                 n_workers=1,
                 use_cgs_p=True
                 ):
        self.n_components = None
        self.topic_word_prior = topic_word_prior
        self.alpha = alpha
        self.epsilon = 1e-20
        self.n_vocabs = None
        self.docstate_ = None
        self.components_ = None
        self.predictor = None
        self.n_workers = n_workers
        self.epsilon = epsilon
        self.add_dummy_topic = add_dummy_topic
        self.n_iter = n_iter
        self.use_cgs_p = use_cgs_p

    def fit(self, X, Y):
        self._fit(X, Y)
        return self

    def fit_transform(self, X, Y, **kwargs):
        result = self._fit(X, **kwargs) + self.doc_topic_prior[np.newaxis, :]
        result /= result.sum(axis=1)[:, np.newaxis]
        return result

    def _fit(self, X, Y, ll_freq=10):
        if not sps.issparse(Y):
            Y = sps.csr_matrix(Y).astype(IntegerType)
        else:
            Y = Y.astype(IntegerType)

        self.n_components = Y.shape[1]
        ones_topic = np.ones(self.n_components, dtype=RealType)
        self.topic_word_prior = number_to_array(
            X.shape[1], 1 / float(self.n_components),
            self.topic_word_prior
        )

        try:
            count, dix, wix = check_array(X)
        except:
            print('Check for X failed.')
            raise

        doc_topic = np.zeros(
            (X.shape[0], self.n_components), dtype=IntegerType)
        word_topic = np.zeros(
            (X.shape[1], self.n_components), dtype=IntegerType)
        topic_counts = np.zeros(self.n_components, dtype=IntegerType)

        docstate = LabelledLDATrainer(
            self.alpha,
            self.epsilon,
            Y,
            count, dix, wix, self.n_components, 42,
            self.n_workers
        )
        docstate.initialize(word_topic, doc_topic, topic_counts)

        with tqdm(range(self.n_iter)) as pbar:
            for _ in pbar:
                docstate.iterate_gibbs(
                    self.topic_word_prior,
                    doc_topic,
                    word_topic,
                    topic_counts
                )

        doc_topic_prior = (
            self.alpha * np.ones(self.n_components, dtype=RealType))

        self.components_ = word_topic.transpose()
        predictor = Predictor(self.n_components, doc_topic_prior, 42)
        if self.use_cgs_p:
            phi = docstate.obtain_phi(
                self.topic_word_prior,
                doc_topic,
                word_topic,
                topic_counts
            )
        else:
            phi = word_topic + self.topic_word_prior[:, np.newaxis]
            phi /= phi.sum(axis=0)[np.newaxis, :]
            phi = phi.transpose()
        predictor.add_beta(phi.transpose())
        self.predictor = predictor

        return doc_topic

    @property
    def phi(self):
        return self.predictor.phis[0]



from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import sys
import setuptools
import os

__version__ = "0.2.2.0"
install_requires = ["pybind11>=2.5", "numpy >= 1.11", "tqdm", "scipy>=1.0.0"]

eigen_include_dir = os.environ.get("EIGEN3_INCLUDE_DIR", None)
if eigen_include_dir is None:
    install_requires.append("requests")


class get_eigen_include(object):
    EIGEN3_URL = "https://gitlab.com/libeigen/eigen/-/archive/3.3.7/eigen-3.3.7.zip"
    EIGEN3_DIRNAME = "eigen-3.3.7"

    def __str__(self):
        if eigen_include_dir is not None:
            return eigen_include_dir

        basedir = os.path.dirname(__file__)
        target_dir = os.path.join(basedir, self.EIGEN3_DIRNAME)
        print(target_dir)
        if os.path.exists(target_dir):
            return target_dir

        download_target_dir = os.path.join(basedir, "eigen3.zip")
        import requests
        import zipfile

        response = requests.get(self.EIGEN3_URL, stream=True)
        with open(download_target_dir, "wb") as ofs:
            for chunk in response.iter_content(chunk_size=1024):
                ofs.write(chunk)

        with zipfile.ZipFile(download_target_dir) as ifs:
            ifs.extractall()

        return target_dir


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


ext_modules = [
    Extension(
        "lda11._lda",
        [
            "src/wrapper.cpp",
            "src/predictor.cpp",
            "src/trainer_base.cpp",
            "src/trainer.cpp",
            "src/child_worker.cpp",
            "src/labelled_lda.cpp",
        ],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            get_eigen_include(),
        ],
        language="c++",
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """
    flags = ["-std=c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError("Unsupported compiler -- at least C++11 support " "is needed!")


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {
        "msvc": ["/EHsc"],
        "unix": ["-march=native"],
    }
    l_opts = {
        "msvc": [],
        "unix": [],
    }

    if sys.platform == "darwin":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.7"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


setup(
    name="lda11",
    version=__version__,
    author="Tomoki Ohtsuki",
    url="https://github.com/tohtsky/lda11",
    author_email="tomoki.ohtsuki129@gmail.com",
    description="Yet another CGS sampler for Latent Dirichlet Allocation.",
    long_description="",
    ext_modules=ext_modules,
    install_requires=install_requires,
    setup_requires=install_requires,
    cmdclass={"build_ext": BuildExt},
    packages=["lda11"],
    zip_safe=False,
)
